{
    "quota_set": {
        "backup_gigabytes": {
            "in_use": 0,
            "limit": 1000,
            "reserved": 0
        },
        "backups": {
            "in_use": 0,
            "limit": 10,
            "reserved": 0
        },
        "gigabytes": {
            "in_use": 0,
            "limit": 1000,
            "reserved": 0
        },
        "gigabytes___DEFAULT__": {
            "in_use": 0,
            "limit": -1,
            "reserved": 0
        },
        "groups": {
            "in_use": 0,
            "limit": 10,
            "reserved": 0
        },
        "id": "fake_tenant",
        "per_volume_gigabytes": {
            "in_use": 0,
            "limit": -1,
            "reserved": 0
        },
        "snapshots": {
            "in_use": 0,
            "limit": 10,
            "reserved": 0
        },
        "snapshots___DEFAULT__": {
            "in_use": 0,
            "limit": -1,
            "reserved": 0
        },
        "volumes": {
            "in_use": 0,
            "limit": 10,
            "reserved": 0
        },
        "volumes___DEFAULT__": {
            "in_use": 0,
            "limit": -1,
            "reserved": 0
        }
    }
}