{
    "volumes": [
        {
            "attachments": [],
            "availability_zone": "nova",
            "bootable": "false",
            "consistencygroup_id": null,
            "created_at": "%(strtime)s",
            "description": null,
            "encrypted": false,
            "id": "%(uuid)s",
            "links": [
                {
                    "href": "%(host)s/v3/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/volumes/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/89afd400-b646-4bbc-b12b-c0a4d63e5bd3/volumes/%(uuid)s",
                    "rel": "bookmark"
                }
            ],
            "metadata": {},
            "migration_status": null,
            "multiattach": false,
            "name": null,
            "os-vol-host-attr:host": null,
            "os-vol-mig-status-attr:migstat": null,
            "os-vol-mig-status-attr:name_id": null,
            "os-vol-tenant-attr:tenant_id": "%(uuid)s",
            "replication_status": null,
            "size": 10,
            "snapshot_id": null,
            "source_volid": null,
            "status": "creating",
            "updated_at": null,
            "user_id": "%(uuid)s",
            "volume_type": "%(name)s",
            "volume_type_id": "%(uuid)s",
            "service_uuid": null,
            "provider_id": null,
            "group_id": null,
            "shared_targets": true,
            "cluster_name": null,
            "consumes_quota": true
        }
    ]
}
