# Stubs for oslo_i18n.fixture (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

import fixtures
import gettext
from typing import Any, Optional

class Translation(fixtures.Fixture):
    domain: Any = ...
    def __init__(self, domain: str = ...) -> None: ...
    def lazy(self, msg: Any): ...
    def immediate(self, msg: Any): ...

class ToggleLazy(fixtures.Fixture):
    def __init__(self, enabled: Any) -> None: ...
    def setUp(self) -> None: ...

class _PrefixTranslator(gettext.NullTranslations):
    prefix: Any = ...
    def __init__(self, fp: Optional[Any] = ..., prefix: str = ...) -> None: ...
    def gettext(self, message: Any): ...
    def ugettext(self, message: Any): ...

class PrefixLazyTranslation(fixtures.Fixture):
    languages: Any = ...
    locale: Any = ...
    def __init__(self, languages: Optional[Any] = ..., locale: Optional[Any] = ...) -> None: ...
    def setUp(self): ...
