# Stubs for oslo_i18n._message (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

import six
from typing import Any, Optional

CONTEXT_SEPARATOR: str
LOG: Any

class Message(six.text_type):
    def __new__(cls, msgid: Any, msgtext: Optional[Any] = ..., params: Optional[Any] = ..., domain: str = ..., has_contextual_form: bool = ..., has_plural_form: bool = ..., *args: Any): ...
    def translation(self, desired_locale: Optional[Any] = ...): ...
    def __mod__(self, other: Any): ...
    def __add__(self, other: Any) -> None: ...
    def __radd__(self, other: Any): ...
