{
    "snapshots": [
        {
            "created_at": "%(strtime)s",
            "description": "Daily backup",
            "id": "%(uuid)s",
            "metadata": {
                "key": "v3"
            },
            "name": "snap-001",
            "os-extended-snapshot-attributes:progress": "0%",
            "os-extended-snapshot-attributes:project_id": "%(uuid)s",
            "size": 10,
            "status": "creating",
            "updated_at": null,
            "volume_id": "%(uuid)s"
        }
    ]
}