{
    "volume_type": {
        "name": "vol-type-001",
        "extra_specs": {
            "capabilities": "gpu"
        },
        "os-volume-type-access:is_public": true,
        "is_public": true,
        "id": "%(uuid)s",
        "description": "volume type 0001"
    }
}
