{
    "encryption": {
        "volume_type_id": "%(uuid)s",
        "control_location": "front-end",
        "encryption_id": "%(uuid)s",
        "key_size": 256,
        "provider": "luks",
        "cipher": "aes-xts-plain64"
    }
}
