# Copyright 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from cinder.api.validation import parameter_types


create = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'qos_specs': {
            'type': 'object',
            'properties': {
                'name': {
                    'type': 'string',
                    'format': 'name_skip_leading_trailing_spaces'
                },
            },
            'required': ['name'],
            'additionalProperties': True,
        },
    },
    'required': ['qos_specs'],
    'additionalProperties': False,
}


set = {
    'type': 'object',
    'properties': {
        'qos_specs': parameter_types.extra_specs_with_null
    },
    'required': ['qos_specs'],
    'additionalProperties': False,
}


unset = {
    'type': 'object',
    'properties': {
        'keys': {
            'type': 'array',
            'items': {
                'type': 'string', 'minLength': 1, 'maxLength': 255
            },
            'uniqueItems': True
        },
    },
    'required': ['keys'],
    'additionalProperties': False,
}
