# Stubs for oslo_i18n._gettextutils (Python 3)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

def install(domain: Any) -> None: ...
def get_available_languages(domain: Any): ...
